;; -*- lexical-binding: t; -*-



;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                        ;;
;;  Affichage de la pile  ;;
;;                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun mon-format (val)
  (if (floatp val)
      (format "%.2f" val)
    (format "%s" val)))


(defun afficher-pile (liste)
  (let ((taille-max (max 10
                         (or (cl-loop for val in liste
                                      maximize (length (mon-format val)))
                             0))))

    (cl-loop
     initially (princ (concat "   ┌─" (string-pad  "─" taille-max ?─) "─┐\n"))
     for val in (reverse liste)
     for chaîne = (mon-format val)
     do (princ (format (concat "   │ "
                               (string-pad chaîne taille-max ?   t) " │\n")))
     finally (princ (concat "   │ " (string-pad  " " taille-max ? ) " │\n"))
     )))





;;;;;;;;;;;;;;;;;;;;;;;;
;;                    ;;
;;  Analyse lexicale  ;;
;;                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;



(defun numbers-p (&rest objets)
  (cl-loop for x in objets
           always (numberp x)))


(defun lire (ch)
  (or (when (equal "#" ch) (list :contrôle "#"))
   (condition-case err
       (let ((token (read ch)))
         (cl-case token

           (/   (list :opération #'diviser))
           (+   (list :opération #'somme))
           (*   (list :opération #'produit))
           (-   (list :opération #'différence))
           (^   (list :opération #'puissance))
           (div (list :opération #'quotient))
           (mod (list :opération #'modulo))

           (++  (list :opération* #'somme*))
           (**  (list :opération* #'produit*))
           (moy (list :opération* #'moyenne*))

           (!   (list :fonction #'fact))
           (rac (list :fonction #'racine))


           (pi    (list :nombre pi))
           (e     (list :nombre (exp 1)))

           (t (if (numberp token)
                  (list :nombre token)
                (list :erreur)))))
     (invalid-read-syntax (message "Erreur de syntaxe") nil)
     (end-of-file (message "Erreur de syntaxe") nil))
   (list :erreur)))

(brouillon
 (lire "\"")
 (lire "1"))


;;;;;;;;;;;;;;;;;;
;;              ;;
;;  Opérations  ;;
;;              ;;
;;;;;;;;;;;;;;;;;;

(defmacro ok (test &rest body)
  (declare (indent defun))
  `(when ,test
     (list :ok (progn ,@body))))




(defun diviser (a b)
  (ok (and (numbers-p a b)
           (/= 0 b))
    (if (and (integerp a)
             (integerp b)
             (= 0 (mod a b)))
        (/ a b)
      (/ (float a)
         (float b)))))

(defun somme (a b)
  (ok (numbers-p a b)
    (+ a b)))

(defun produit (a b)
  (ok (numbers-p a b)
    (* a b)))

(defun différence (a b)
  (ok (numbers-p a b)
    (- a b)))

(defun puissance (a b)
  (condition-case err
      (ok (and (numbers-p a b)
               (or (/= a 0)
                   (<= 0 b)))
        (expt a b))
    (overflow-error (message "Erreur : nombre trop grand") nil)))

(defun quotient (a b)
  (ok (and (integerp a)
           (integerp b)
           (/= 0 b))
    (/ a b)))

(defun modulo (a b)
  (ok (and (integerp a)
           (integerp b)
           (/= 0 b))
    (mod a b)))


(defun somme* (liste)
  (ok (apply #'numbers-p liste)
    (apply  #'+ liste)))

(defun produit* (liste)
  (ok (apply #'numbers-p liste)
    (apply  #'* liste)))


(defun moyenne* (liste)
  (ok (apply #'numbers-p liste)
    (/ (apply  #'+ liste)
       (float (length liste)))))






;;;;;;;;;;;;;;;;;;
;;              ;;
;;  Fonctions   ;;
;;              ;;
;;;;;;;;;;;;;;;;;;

(defun racine (x)
  (ok (<= 0 x)
    (expt x 0.5)))

(defun fact (n)
  (ok (and (integerp n)
           (<= 0 n))
    (cl-loop for i from 0 to n
             for p = 1 then (* p i)
             finally (return p))))



;;;;;;;;;;;;;;;;;;
;;              ;;
;;  Évaluateur  ;;
;;              ;;
;;;;;;;;;;;;;;;;;;

(defmacro pop* (liste)
    `(cl-loop while ,liste
              for x = (pop ,liste)
              until (equal x "#")
              collect x))

(defmacro push* (liste pile)
  `(dolist (x (cons "#" (reverse ,liste)))
     (push x ,pile)))

(brouillon
 (let ((L (list 4 5 6)))
  (push* (list 1 2 3) L )
  L)

 (let ((L (list 1 2 3 4 5 "#" 6 7 8 9)))
  (concat (format "%s\n" (pop* L))
          (format "%s\n" L)))


 (let ((L (list 1 2 3 4 5 6 7 8 9)))
  (concat (format "%s\n" (pop* L))
          (format "%s\n" L)))

 nil)


(defun exécute (liste)
  (let (pile)
    (dolist (val liste)
      (unless (eq 'erreur (car pile))

        (cl-case (car val)
          (:opération
           (if (<= 2 (length pile))
               (let* ((a        (pop pile))
                      (b        (pop pile))
                      (v        (cadr val))
                      (résultat (funcall v b a)))

                 (if résultat
                     (push (cadr résultat) pile)
                   (push b pile)
                   (push a pile)
                   (push 'erreur pile)))
             (push 'erreur pile)))


          (:opération*
           (let* ((liste    (pop* pile))
                  (v        (cadr val))
                  (résultat (funcall v liste)))

             (if résultat
                 (push (cadr résultat) pile)
               (push* liste pile)
               (push 'erreur pile))))



          (:fonction
           (if (<= 1 (length pile))
               (let* ((a        (pop pile))
                      (v        (cadr val))
                      (résultat (funcall v a)))

                 (if résultat
                     (push (cadr résultat) pile)
                   (push a pile)
                   (push 'erreur pile)))
             (push 'erreur pile)))

          (:contrôle
           (push (cadr val) pile))

          (:nombre
           (push (cadr val) pile))

          (:erreur
           (push 'erreur pile)))))
    pile))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                          ;;
;;  Déplacement pour emacs  ;;
;;                          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun espace-p (x)
  (member x (list " " "\n" "\t" "\f")))

(defun prochain-non-vide ()
  (interactive)
  (let ((i 0))
    (while (espace-p (thing-at-point 'char t))
      (message (format "%d" i))
      (right-char))
    (while (not (espace-p (thing-at-point 'char t)))
      (message (format "%d" i))
      (right-char))))

(defun précédent-non-vide ()
  (interactive)
  (right-char -1)
  (while (not (espace-p (thing-at-point 'char t)))
    (right-char -1))
  (while (espace-p (thing-at-point 'char t))
    (right-char -1))
  (right-char 1))

(defun calcule-jusqu-au-point ()
  (interactive)
  (exécute
   (mapcar #'read
           (split-string
            (buffer-substring-no-properties 1 (point))))))



;;;;;;;;;;;;;;;;;;
;;              ;;
;;  Mode Emacs  ;;
;;              ;;
;;;;;;;;;;;;;;;;;;

(defun pile()
  (interactive)
  (switch-to-buffer "*brouillon*")
  (split-window-horizontally)
  (windmove-right)
  (switch-to-buffer "*PILE*")
  (windmove-left)
  (insert "\n\n\t")
  (local-set-key (kbd "C-<right>") 'prochain-non-vide)
  (local-set-key (kbd "C-<left>") 'précédent-non-vide)
  (add-hook 'post-command-hook 'calcule-pile-vers-buffer 0 t))

(defun calcule-pile-vers-buffer ()
  (interactive)
  (with-output-to-temp-buffer "*PILE*"
    (afficher-pile
     (exécute
      (mapcar #'lire
              (split-string
               (buffer-substring-no-properties 1 (point))))))))


(defun calcule-pile-vers-buffer-debug ()
  (interactive)
  (exécute
   (mapcar #'lire
           (split-string
            (buffer-substring-no-properties 1 (point))))))


(defun test-pile (chaîne)
  (exécute
   (mapcar #'lire
           (split-string chaîne))))
